/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.platform.raspberry.generator.extension;

import static org.fortiss.af3.component.generator.c.ComponentProgramToCSourcePackageTransformation.getPortNoValIdentifier;
import static org.fortiss.af3.component.generator.component.PortVariableUtils.getPortIdentifier;
import static org.fortiss.af3.platform.raspberry.generator.templates.RasPiCTemplates.getSingleUnitMainCFile;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.conqat.lib.commons.collections.Pair;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.component.model.InputPort;
import org.fortiss.af3.component.model.OutputPort;
import org.fortiss.af3.component.model.Port;
import org.fortiss.af3.generator.common.model.source.AbstractUnit;
import org.fortiss.af3.platform.language.executable.ExecutableBase;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.fortiss.af3.platform.model.PlatformConnectorUnit;
import org.fortiss.af3.platform.model.Receiver;
import org.fortiss.af3.platform.model.Transmitter;
import org.fortiss.af3.platform.raspberry.generator.framework.IInstanceGeneratorExtension;
import org.fortiss.af3.platform.raspberry.generator.framework.IInstanceInitializationGeneratorExtension;
import org.fortiss.af3.platform.raspberry.generator.framework.IReadableGeneratorExtension;
import org.fortiss.af3.platform.raspberry.generator.framework.ISingletonGeneratorExtension;
import org.fortiss.af3.platform.raspberry.generator.framework.ISingletonInitializationGeneratorExtension;
import org.fortiss.af3.platform.raspberry.generator.framework.IWriteableGeneratorExtension;
import org.fortiss.af3.platform.raspberry.model.RaspberryPi;
import org.fortiss.tooling.kernel.extension.data.ITransformationContext;

/** Separate class for generating the main.c file for deployments ion single {@link RaspberryPi}. */
class SingleUnitMainGenerator {
	/** The current execution unit. */
	private RaspberryPi executionUnit;
	/** The deployed components. */
	private List<Pair<ExecutionUnit, Component>> deployedComponents;
	/** The deployed ports. */
	private List<Pair<PlatformConnectorUnit, Port>> deployedPorts;
	/** The transformation context. */
	private ITransformationContext context;
	/** The set of platform header files to be included. */
	private Set<String> headers;
	/** The map from platform elements to executables. */
	private Map<PlatformConnectorUnit, ExecutableBase<?>> platformConnector2ExecutableBase =
			new HashMap<>();
	/** The map from singleton identifiers to singleton prefixes. */
	private Map<String, String> singletonPostfixMap = new HashMap<>();
	/** The map from executables to instance prefixes. */
	private Map<ExecutableBase<?>, String> instancePostfixMap = new HashMap<>();

	/** Constructor. */
	public SingleUnitMainGenerator(RaspberryPi executionUnit,
			List<Pair<ExecutionUnit, Component>> deployedComponents,
			List<Pair<PlatformConnectorUnit, Port>> deployedPorts,
			Map<PlatformConnectorUnit, ExecutableBase<?>> platformConnector2ExecutableBase,
			Set<String> headers, ITransformationContext context) {
		this.executionUnit = executionUnit;
		this.deployedComponents = deployedComponents;
		this.deployedPorts = deployedPorts;
		this.headers = headers;
		this.context = context;
		this.platformConnector2ExecutableBase.putAll(platformConnector2ExecutableBase);
		computePostfixes();
	}

	/** Creates the main.c file for deployments with a single execution units. */
	public AbstractUnit createSingleUnitMain() {
		String includes = createIncludes();
		String auxFuns = createAuxiliaryFunctions();
		String initCode = createInitCode();
		String sensorVariables = createVariables();
		String workerCode = createWorkerCode(deployedComponents, deployedPorts);
		return getSingleUnitMainCFile(executionUnit.getName(), executionUnit.getCycleTime(),
				includes, initCode, auxFuns, workerCode, sensorVariables);
	}

	/** Creates the code for auxiliary functions. */
	private String createAuxiliaryFunctions() {
		StringBuilder sb = new StringBuilder();
		Set<String> singletonInitDone = new HashSet<>();
		for(PlatformConnectorUnit pcu : platformConnector2ExecutableBase.keySet()) {
			ExecutableBase<?> exec = platformConnector2ExecutableBase.get(pcu);
			String sPostfix = null;
			if(exec instanceof ISingletonInitializationGeneratorExtension) {
				ISingletonInitializationGeneratorExtension singletonExec =
						(ISingletonInitializationGeneratorExtension)exec;
				// check if function definition was already included
				String ident = singletonExec.getSingletonIdentifier();
				sPostfix = singletonPostfixMap.get(ident);
				if(!singletonInitDone.contains(ident)) {
					singletonInitDone.add(ident);
					String def = singletonExec.getSingletonAuxiliaryFunctions(sPostfix);
					sb.append(def != null ? def : "");
				}
			}
			// sPostfix is set by first if-block or remains null
			if(exec instanceof IInstanceInitializationGeneratorExtension) {
				IInstanceInitializationGeneratorExtension initExec =
						(IInstanceInitializationGeneratorExtension)exec;
				String iPostfix = instancePostfixMap.get(initExec);
				String def = initExec.getInstanceAuxiliaryFunctions(iPostfix, sPostfix);
				sb.append(def != null ? def : "");
			}
		}
		return sb.toString();
	}

	/** Creates the includes of the system headers. */
	private String createIncludes() {
		// create includes for logical components
		StringBuilder sb = new StringBuilder();
		for(Pair<ExecutionUnit, Component> p : deployedComponents) {
			Component c = p.getSecond();
			sb.append("#include <" + c.getName() + "_ID_" + c.getId() + ".h>\n");
		}
		// create includes for platform header files
		for(String header : headers) {
			sb.append("#include <" + header + ">\n");
		}
		// create standard includes
		sb.append("#include <stdio.h>\n");
		return sb.toString();
	}

	/** Create the worker code. */
	private String createWorkerCode(List<Pair<ExecutionUnit, Component>> deployedComponents,
			List<Pair<PlatformConnectorUnit, Port>> deployedPorts) {
		StringBuilder sb = new StringBuilder();
		// if(useUS || useLaser || useACC) {
		// // Get Current time for the sensors
		// sb.append("uint64_t curr_time = time_util_get_current_micros();\n");
		// }
		for(Pair<PlatformConnectorUnit, Port> p : deployedPorts) {
			if(p.getFirst() instanceof Receiver && p.getSecond() instanceof InputPort) {
				sb.append(createReadCode((Receiver)p.getFirst(), (InputPort)p.getSecond()));
			}
		}
		for(Pair<ExecutionUnit, Component> p : deployedComponents) {
			sb.append(makeCall("perform_step", p.getSecond()));
		}
		for(Pair<PlatformConnectorUnit, Port> p : deployedPorts) {
			if(p.getFirst() instanceof Transmitter && p.getSecond() instanceof OutputPort) {
				sb.append(createWriteCode((Transmitter)p.getFirst(), (OutputPort)p.getSecond()));
			}
		}
		return sb.toString();
	}

	/** Create the Sensor Variables and functions. */
	private String createVariables() {
		StringBuilder sb = new StringBuilder();
		Set<String> singletonInitDone = new HashSet<>();
		for(PlatformConnectorUnit pcu : platformConnector2ExecutableBase.keySet()) {
			ExecutableBase<?> exec = platformConnector2ExecutableBase.get(pcu);
			String sPostfix = null;
			if(exec instanceof ISingletonInitializationGeneratorExtension) {
				ISingletonInitializationGeneratorExtension singletonExec =
						(ISingletonInitializationGeneratorExtension)exec;
				// check if variable declaration was already included
				String ident = singletonExec.getSingletonIdentifier();
				sPostfix = singletonPostfixMap.get(ident);
				if(!singletonInitDone.contains(ident)) {
					singletonInitDone.add(ident);
					String def = singletonExec.getSingletonVariableDeclarationCode(sPostfix);
					sb.append(def != null ? def : "");
				}
			}
			// sPostfix is set by first if-block or remains null
			if(exec instanceof IInstanceInitializationGeneratorExtension) {
				IInstanceInitializationGeneratorExtension initExec =
						(IInstanceInitializationGeneratorExtension)exec;
				String iPostfix = instancePostfixMap.get(initExec);
				String def = initExec.getVariableDeclaration(iPostfix, sPostfix);
				sb.append(def != null ? def : "");
			}
		}
		return sb.toString();
	}

	// if(useDigits) {
	// sb.append("SegmentDisplay4x7 segment_display;\n");
	// sb.append("static void set_led_display (SegmentDisplay4x7 *segment_display, uint16_t value, bool showHex) {\n");
	// sb.append("static const uint8_t digits[] = {0x3f,0x06,0x5b,0x4f,\n");
	// sb.append("                                 0x66,0x6d,0x7d,0x07,\n");
	// sb.append("                                 0x7f,0x6f,0x77,0x7c,\n");
	// sb.append("                                 0x39,0x5e,0x79,0x71};\n");
	// sb.append("if(showHex) {\n");
	// sb.append("uint8_t segments[4] = {digits[(value >> 12) & 0x0F], digits[(value >> 8) & 0x0F], digits[(value >> 4) & 0x0F], digits[value & 0x0F]};\n");
	// sb.append("segment_display_4x7_set_segments(segment_display, segments, 5, false);\n");
	// sb.append("} else {\n");
	// sb.append("uint8_t segments[4] = {digits[(value%10000)/1000], digits[(value%1000)/100], digits[(value%100)/10], digits[value%10]};\n");
	// sb.append("segment_display_4x7_set_segments(segment_display, segments, 5, false);\n");
	// sb.append("}\n");
	// sb.append("}\n\n");
	// }
	// if(useLEDButton) {
	// // for(Pair<PlatformConnectorUnit, Port> p : deployedPorts) {
	// for(String uid : ledButtons.keySet()) {
	// sb.append("RGBLEDButton rlb_" + uid + ";\n");
	// }
	// for(String[] arr : ledButtonsLocalVariable.values()) {
	// for(String var : arr) {
	// sb.append("uint8_t " + var + " = 0;\n");
	// }
	// }
	// // extractRGBforLEDButton(p);
	// // }
	// }
	// if(useACC) {
	// sb.append("float Q = 0.022;\n");
	// sb.append("float R = 0.917;\n");
	// sb.append("float estimates[9] = {0.0, 0.0, 0.0, 0.0, 0.0, 0.0};\n");
	// sb.append("float last_p[9] = {0.0, 0.0, 0.0, 0.0, 0.0, 0.0};\n");
	// sb.append("float kalman_filter(float value, int sensor) {\n");
	// sb.append("if(sensor < 0 || sensor > 5) return -1;\n");
	// sb.append("float temp_est = estimates[sensor];\n");
	// sb.append("float p_temp = last_p[sensor] + Q;\n");
	// sb.append("float K = p_temp * (1.0/(p_temp + R));\n");
	// sb.append("float est = temp_est + K * (value - temp_est);\n");
	// sb.append("float P = (1-K)*p_temp;\n");
	// sb.append("last_p[sensor] = P;\n");
	// sb.append("estimates[sensor] = est;\n");
	// sb.append("return est;\n");
	// sb.append("}\n\n");
	// sb.append("float acceleration_X, acceleration_Y, acceleration_Z;\n");
	// sb.append("uint64_t acc_last_cb_time = 0;\n");
	// sb.append("void acc_callback(int16_t x, int16_t y, int16_t z, void* data) {\n");
	// sb.append("if (abs(x - acceleration_X) > 2 * 10 * 100 ||\n");
	// sb.append("abs(y - acceleration_Y) > 2 * 10 * 100 ||\n");
	// sb.append("abs(z - acceleration_Z) > 2 * 10 * 100)\n");
	// sb.append("    return;\n");
	// sb.append("acceleration_X = kalman_filter(x, 0) / 100.0;\n");
	// sb.append("acceleration_Y = kalman_filter(y, 1) / 100.0;\n");
	// sb.append("acceleration_Z = kalman_filter(z, 2) / 100.0;\n");
	// sb.append("acc_last_cb_time = time_util_get_current_micros();\n");
	// sb.append("}\n\n");
	// sb.append("float angularVelocity_X, angularVelocity_Y, angularVelocity_Z;\n");
	// sb.append("uint64_t angV_last_cb_time = 0;\n");
	// sb.append("void angV_callback(int16_t x, int16_t y, int16_t z, void* data) {\n");
	// sb.append("angularVelocity_X = kalman_filter(x, 3) / 16;\n");
	// sb.append("angularVelocity_Y = kalman_filter(y, 4) / 16;\n");
	// sb.append("angularVelocity_Z = kalman_filter(z, 5) / 16;\n");
	// sb.append("angV_last_cb_time = time_util_get_current_micros();\n");
	// sb.append("}\n\n");
	// }

	/**
	 * @param p
	 */
	// private void extractRGBforLEDButton(Pair<PlatformConnectorUnit, Port> p) {
	// final String UNKNOWN = "unknown";
	// String uid = UNKNOWN;
	// int rgb = -1;
	// if(p.getFirst() instanceof LEDButtonRActuator) {
	// LEDButtonRActuator sensor = (LEDButtonRActuator)p.getFirst();
	// uid = sensor.getUniqueBrickletID();
	// rgb = 0;
	// }
	// if(p.getFirst() instanceof LEDButtonGActuator) {
	// LEDButtonGActuator sensor = (LEDButtonGActuator)p.getFirst();
	// uid = sensor.getUniqueBrickletID();
	// rgb = 1;
	// }
	// if(p.getFirst() instanceof LEDButtonBActuator) {
	// LEDButtonBActuator sensor = (LEDButtonBActuator)p.getFirst();
	// uid = sensor.getUniqueBrickletID();
	// rgb = 2;
	// }
	// if(!uid.equals(UNKNOWN) && !ledButtons.containsKey(uid)) {
	// String[] rgbArray = new String[3];
	// rgbArray[rgb] = portName(p.getSecond());
	// ledButtons.put(uid, rgbArray);
	// String[] localVar = new String[3];
	// localVar[rgb] =
	// "rgbButton_" + uid + "_lastValue_" + (rgb == 0 ? "r" : rgb == 1 ? "g" : "b");
	// ledButtonsLocalVariable.put(uid, localVar);
	// } else if(ledButtons.containsKey(uid)) {
	// String[] rgbArray = ledButtons.get(uid);
	// rgbArray[rgb] = portName(p.getSecond());
	// String[] localVar = ledButtonsLocalVariable.get(uid);
	// localVar[rgb] =
	// "rgbButton_" + uid + "_lastValue_" + (rgb == 0 ? "r" : rgb == 1 ? "g" : "b");
	// }
	// }

	/** Creates the read code for the given receiver and port. */
	private String createWriteCode(Transmitter transmitter, OutputPort outport) {
		ExecutableBase<?> exec = platformConnector2ExecutableBase.get(transmitter);
		String sPostfix = null;
		if(exec instanceof ISingletonGeneratorExtension) {
			sPostfix =
					singletonPostfixMap.get(((ISingletonGeneratorExtension)exec)
							.getSingletonIdentifier());
		}
		String iPostfix = instancePostfixMap.get(exec);
		if(exec instanceof IWriteableGeneratorExtension) {
			IWriteableGeneratorExtension wexec = (IWriteableGeneratorExtension)exec;
			String sourceVariable = getPortIdentifier(outport);
			String writeCode = wexec.getWriteCode(iPostfix, sPostfix, outport, sourceVariable);
			String novalCode = wexec.getNoValWriteCode(iPostfix, sPostfix, outport);
			if(novalCode != null) {
				String novalVariable = getPortNoValIdentifier(sourceVariable);
				return "if (" + novalVariable + ") {\n" + novalCode + "} else {\n" + writeCode +
						"}\n";
			}
			return writeCode;
		}
		return "FIXME(\"No generator extension available to write " + outport.getName() + " to " +
				transmitter.getName() + "\");\n";
	}

	/** Creates the read code for the given receiver and port. */
	private String createReadCode(Receiver receiver, InputPort inport) {
		ExecutableBase<?> exec = platformConnector2ExecutableBase.get(receiver);
		String sPostfix = null;
		if(exec instanceof ISingletonGeneratorExtension) {
			sPostfix =
					singletonPostfixMap.get(((ISingletonGeneratorExtension)exec)
							.getSingletonIdentifier());
		}
		String iPostfix = instancePostfixMap.get(exec);
		if(exec instanceof IReadableGeneratorExtension) {
			IReadableGeneratorExtension rexec = (IReadableGeneratorExtension)exec;
			String targetVariable = getPortIdentifier(inport);
			String readCode = rexec.getReadCode(iPostfix, sPostfix, inport, targetVariable);
			String novalCode = rexec.getNoValReadCode(iPostfix, sPostfix, inport);
			if(novalCode != null) {
				return "if (" + novalCode + ") {\n" + readCode + "}\"";
			}
			return readCode;
		}
		return "FIXME(\"No generator extension available to read " + inport.getName() + " from " +
				receiver.getName() + "\");\n";
	}

	/** Create the initialize code. */
	private String createInitCode() {
		StringBuilder sb = new StringBuilder();
		Set<String> singletonInitDone = new HashSet<>();
		for(PlatformConnectorUnit pcu : platformConnector2ExecutableBase.keySet()) {
			ExecutableBase<?> exec = platformConnector2ExecutableBase.get(pcu);
			String sPostfix = null;
			if(exec instanceof ISingletonInitializationGeneratorExtension) {
				ISingletonInitializationGeneratorExtension singletonExec =
						(ISingletonInitializationGeneratorExtension)exec;
				String ident = singletonExec.getSingletonIdentifier();
				sPostfix = singletonPostfixMap.get(ident);
				if(!singletonInitDone.contains(ident)) {
					singletonInitDone.add(ident);
					String code = singletonExec.getSingletonInitializationCode(sPostfix);
					sb.append(code != null ? code : "");
				}
			}
			// sPostfix is set in the first if-block or is null
			if(exec instanceof IInstanceInitializationGeneratorExtension) {
				IInstanceInitializationGeneratorExtension initExec =
						(IInstanceInitializationGeneratorExtension)exec;
				String iPostfix = instancePostfixMap.get(initExec);
				String code = initExec.getInitializationCode(iPostfix, sPostfix);
				sb.append(code != null ? code : "");
			}
		}
		return sb.toString();
	}

	/** A counter to produce unique identifiers. */
	private int postfixCounter = 1;

	/** Computes the singleton and instance prefixes for all executables. */
	private void computePostfixes() {
		for(PlatformConnectorUnit pcu : platformConnector2ExecutableBase.keySet()) {
			ExecutableBase<?> exec = platformConnector2ExecutableBase.get(pcu);
			if(exec instanceof ISingletonGeneratorExtension) {
				String id = ((ISingletonGeneratorExtension)exec).getSingletonIdentifier();
				if(!singletonPostfixMap.containsKey(id)) {
					String sPostfix = "singleton_" + (postfixCounter++);
					singletonPostfixMap.put(id, sPostfix);
				}
			}
			if(exec instanceof IInstanceGeneratorExtension) {
				if(!instancePostfixMap.containsKey(exec)) {
					String iPostfix = "instance_" + (postfixCounter++);
					instancePostfixMap.put(exec, iPostfix);
				}
			}
		}
	}

	private void temp() {
		// int createdUsSensors = 0;
		// StringBuilder sb = new StringBuilder();
		// if(useCamera) {
		// String addr = executionUnit.getCameraServerAddress();
		// String port = executionUnit.getCameraServerPort();
		// sb.append("camera_client_initialize(\"" + addr + "\", \"" + port + "\");\n");
		// }
		// if(useDigits) {
		// for(Pair<PlatformConnectorUnit, Port> p : deployedPorts) {
		// if(p.getFirst() instanceof ActuatorDigits) {
		// ActuatorDigits sensor = (ActuatorDigits)p.getFirst();
		// sb.append("segment_display_4x7_create(&segment_display, \"" +
		// sensor.getUniqueBrickletID() + "\", &brick_connection);\n");
		// break;
		// }
		// }
		// }
		// if(useLEDButton) {
		// for(String uid : ledButtons.keySet()) {
		// sb.append("rgb_led_button_create(&rlb_" + uid + ", \"" + uid +
		// "\", &brick_connection);\n");
		// }
		// // break; TODO(JE): those break statements only allow one element of each type!
		// }
		// if(useACC) {
		// sb.append("IMUV2 imu;\n");
		// for(Pair<PlatformConnectorUnit, Port> p : deployedPorts) {
		// if(p.getFirst() instanceof AccelerationXSensor ||
		// p.getFirst() instanceof AccelerationYSensor ||
		// p.getFirst() instanceof AccelerationZSensor ||
		// p.getFirst() instanceof AngularVelocityXSensor ||
		// p.getFirst() instanceof AngularVelocityYSensor ||
		// p.getFirst() instanceof AngularVelocityZSensor) {
		// UIDUnit sensor = (UIDUnit)p.getFirst();
		// sb.append("imu_v2_create(&imu, \"" + sensor.getUniqueBrickletID() +
		// "\", &brick_connection);\n");
		// break;
		// }
		// }
		// sb.append("imu_v2_register_callback(&imu, IMU_V2_CALLBACK_LINEAR_ACCELERATION, (void*)acc_callback, NULL);\n");
		// sb.append("imu_v2_set_linear_acceleration_period(&imu, 5);\n");
		// sb.append("imu_v2_register_callback(&imu, IMU_V2_CALLBACK_ANGULAR_VELOCITY, (void*)angV_callback, NULL);\n");
		// sb.append("imu_v2_set_angular_velocity_period(&imu, 5);\n");
		// }
		// if(usePWM) {
		// sb.append("maestro_fd = temp_actuator_initialize(\"/dev/Maestro0\");\n");
		// }
		// if(useGamepad) {
		// sb.append("gamepad_configuration_t* gamepad_config = malloc(sizeof(gamepad_configuration_t));\n");
		// sb.append("gamepad_config->device_id = \"/dev/input/js0\";\n");
		// sb.append("gamepad_config->waiting_sleep_in_micros = " + WAITING_SLEEP_IN_MICROS +
		// ";\n");
		// sb.append("gamepad_config->axis_callback = NULL;\n");
		// sb.append("gamepad_config->button_callback = NULL;\n");
		// sb.append("gamepad_initialize(gamepad_config);\n\n");
		// }
		// if(useRumblepad) {
		// sb.append("while(access(\"/dev/input/event0\", F_OK) == -1) {\n");
		// sb.append("sleep(1);\n");
		// sb.append("printf(\"Could not access gamepad device. Trying again in 1s.\\n\");\n");
		// sb.append("}\n");
		// sb.append("rumblepad_configuration_t* rumblepad_config = malloc(sizeof(rumblepad_configuration_t));\n");
		// sb.append("rumblepad_config->device_id = \"/dev/input/event0\";\n");
		// sb.append("rumblepad_config->waiting_sleep_in_micros = " + WAITING_SLEEP_IN_MICROS +
		// ";\n");
		// sb.append("rumblepad_config->axis_callback = NULL;\n");
		// sb.append("rumblepad_config->button_callback = NULL;\n");
		// sb.append("rumblepad_initialize(rumblepad_config);\n\n");
		// }
		// for(Pair<ExecutionUnit, Component> p : deployedComponents) {
		// Component c = p.getSecond();
		// sb.append(makeCall("init", c));
		// }
		// return sb.toString();
	}

	/** Creates the function call code. */
	private String makeCall(String function, Component c) {
		StringBuilder sb = new StringBuilder();
		sb.append(function).append('_').append(c.getName());
		sb.append("_ID_").append(c.getId()).append("();\n");
		return sb.toString();
	}
}
